
from pathlib import Path

from click import ClickException
from click import command
from click import option
from click import secho
from click import version_option

from versionoverlord import __version__

from versionoverlord.Common import AdvancedSlugs
from versionoverlord.Common import CLISlugs
from versionoverlord.Common import EPILOG
from versionoverlord.Common import extractCLISlugs

from versionoverlord.Common import setUpLogging

from versionoverlord.FileNameToSlugs import FileNameToSlugs
from versionoverlord.TemplateHandler import TemplateHandler

from versionoverlord.githubadapter.GitHubAdapterError import GitHubAdapterError


# noinspection SpellCheckingInspection
@command(epilog=EPILOG)
@version_option(version=f'{__version__}', message='%(prog)s version %(version)s')
@option('--slugs',     '-s',  multiple=True, required=False, help='Create package update specification')
@option('--input-file', '-i', required=False,                help='Use input file for slug list')
def createSpecification(slugs: CLISlugs, input_file: str):
    """
    \b
    This command creates .csv specification file
    It uses the following environment variables:
    \b
        GH_TOKEN      – A personal GitHub access token necessary to read repository release information
        PROJECTS_BASE – The local directory where the python projects are based
        PROJECT – The name of the project;  It should be a directory name
    """

    try:
        if len(slugs) != 0:
            cliSlugs:        AdvancedSlugs   = extractCLISlugs(slugs=slugs)
            templateHandler: TemplateHandler = TemplateHandler(cliSlugs)

            templateHandler.createSpecification()
        elif input_file is not None:
            fqFileName: Path = Path(input_file)
            if fqFileName.exists() is False:
                secho('                          ', fg='red', bg='black', bold=True)
                secho('Input file does not exist ', fg='red', bg='black', bold=True)
                secho('                          ', fg='red', bg='black', bold=True)
            else:
                fileNameToSlugs: FileNameToSlugs = FileNameToSlugs(path=fqFileName)
                fileSlugs:       AdvancedSlugs   = fileNameToSlugs.getSlugs()
                handler:         TemplateHandler = TemplateHandler(advancedSlugs=fileSlugs)

                handler.createSpecification()
    except GitHubAdapterError as e:
        raise ClickException(message=e.message)


if __name__ == "__main__":
    setUpLogging()
    # noinspection SpellCheckingInspection
    # createSpecification(['-i', 'tests/resources/testdata/query.slg'])
    # createSpecification(['-s', 'hasii2011/code-ally-basic,codeallybasic'])
    # -s hasii2011/ -s hasii2011/buildlackey

    createSpecification(['-i', 'docs/query.slg'])




from typing import NewType

from datetime import date

from dataclasses import dataclass

from semantic_version import Version as SemanticVersion

ReleaseTitle   = NewType('ReleaseTitle', str)
ReleaseId      = NewType('ReleaseId',   int)
ReleaseNumber  = NewType('ReleaseNumber', int)


@dataclass
class AdapterMilestone:
    """
    Synthetic class for github.Milestone.Milestone
    """
    title:       str = ''
    state:       str = ''
    description: str = ''
    dueDate:     date | None = date.today()

    releaseNumber: ReleaseNumber = ReleaseNumber(0)
    milestoneUrl:  str           = ''


@dataclass
class AdapterRelease:
    """
    Synthetic class for GitRelease
    """
    body:  str = ''
    draft: bool = True

    title: ReleaseTitle    = ReleaseTitle('')
    tag:   SemanticVersion = SemanticVersion('0.0.0')
    id:    ReleaseId       = ReleaseId(0)




from typing import Any
from typing import Optional

from logging import Logger
from logging import getLogger

from click import ParamType
from click import Parameter
from click import Context

from semantic_version import Version as SemanticVersion


class TagType(ParamType):
    name: str = 'TagType'

    def __init__(self):
        self.logger: Logger = getLogger(__name__)

    def convert(self, value: Any, param: Optional['Parameter'], ctx: Optional['Context']) -> Any:

        try:
            version: SemanticVersion = SemanticVersion(value)
            return version
        except ValueError:
            self.fail(f"{value!r} is not a valid semantic version", param, ctx)




from logging import Logger
from logging import getLogger

from click import ClickException
from click import secho

from semantic_version import Version as SemanticVersion

from versionoverlord.Common import AdvancedSlugs
from versionoverlord.Common import SlugVersion
from versionoverlord.Common import SlugVersions

from versionoverlord.DisplayVersions import DisplayVersions

from versionoverlord.githubadapter.GitHubAdapter import GitHubAdapter

from versionoverlord.githubadapter.GitHubAdapterError import GitHubAdapterError


class SlugHandler:
    def __init__(self, advancedSlugs: AdvancedSlugs):

        self.logger:         Logger        = getLogger(__name__)
        self._advancedSlugs: AdvancedSlugs = advancedSlugs

    def handleSlugs(self):
        try:
            gitHubAdapter: GitHubAdapter = GitHubAdapter()

            slugVersions: SlugVersions = SlugVersions([])
            for advancedSlug in self._advancedSlugs:

                version:     SemanticVersion = gitHubAdapter.getLatestVersionNumber(advancedSlug.slug)
                slugVersion: SlugVersion     = SlugVersion(slug=advancedSlug.slug, version=str(version))
                slugVersions.append(slugVersion)

            if len(slugVersions) == 0:
                secho('Nothing to see here')
            else:
                displayVersions: DisplayVersions = DisplayVersions()
                displayVersions.displaySlugs(slugVersions=slugVersions)
        except GitHubAdapterError as e:
            raise ClickException(message=e.message)

        except (ValueError, Exception) as e:
            print(f'{e}')




from typing import Dict
from typing import List
from typing import NewType
from typing import cast

from logging import Logger
from logging import getLogger

from dataclasses import dataclass
import csv

from pathlib import Path as PyPath

from click import Path
from click import command
from click import echo
from click import option
from click import secho
from click import version_option

from semantic_version import Version as SemanticVersion

from versionoverlord import __version__
from versionoverlord.Common import EPILOG

from versionoverlord.Common import PackageName
from versionoverlord.Common import Packages
from versionoverlord.Common import UpdatePackage
from versionoverlord.Common import runCommand
from versionoverlord.Common import setUpLogging
from versionoverlord.IHandler import IHandler

from versionoverlord.setup.HandleSetupPy import HandleSetupPy

from versionoverlord.circleci.HandleCircleCI import HandleCircleCI

from versionoverlord.pyprojecttoml.HandlePyProjectToml import HandlePyProjectToml

from versionoverlord.requirements.HandleRequirementsTxt import HandleRequirementsTxt


@dataclass
class HandlerSpecification:
    fileName: str      = ''
    handler:  IHandler = cast(IHandler, None)


HandlerSpecifications = NewType('HandlerSpecifications', List[HandlerSpecification])


class UpdateDependencies:
    def __init__(self, specification: PyPath):
        """

        Args:
            specification:  Path to the CSV file
        """
        self.logger: Logger = getLogger(__name__)

        self._packages: Packages = self._buildAPackageList(specification=specification)

        self._handlers: HandlerSpecifications = HandlerSpecifications([
            HandlerSpecification('setup.py',         HandleSetupPy(packages=self._packages)),
            HandlerSpecification('config.yml',       HandleCircleCI(packages=self._packages)),
            HandlerSpecification('requirements.txt', HandleRequirementsTxt(packages=self._packages)),
            HandlerSpecification('pyproject.toml',   HandlePyProjectToml(packages=self._packages))
        ])

    @property
    def packages(self) -> Packages:
        return self._packages

    def update(self):

        assert len(self._packages) != 0,  'Developer error; package list not initialized'

        for spec in self._handlers:
            handlerSpecification: HandlerSpecification = cast(HandlerSpecification, spec)
            handler:              IHandler             = handlerSpecification.handler
            name:                 str                  = handlerSpecification.fileName
            if handler.configurationExists is True:
                # echo(f'Update {name}', color=True)    # This could be misleading if there were no changes
                handler.update()
            else:
                echo(f'No {name}')

    def _buildAPackageList(self, specification: PyPath) -> Packages:
        """
        Transforms the CSV file to a list of Package objects

        Args:
            specification:  The path to the csv file

        Returns:  A list of packages
        """
        with open(specification) as csvfile:
            csvreader = csv.DictReader(csvfile, delimiter=',', quotechar='|')
            packages: Packages = Packages([])

            for dictRow in csvreader:
                row: Dict[str, str] = cast(Dict[str, str], dictRow)
                self.logger.debug(row['PackageName'], row['OldVersion'], row['NewVersion'])
                packageName: PackageName = PackageName(row['PackageName'])
                updatePackage: UpdatePackage = UpdatePackage()
                updatePackage.packageName = packageName
                try:
                    updatePackage.oldVersion = SemanticVersion(row['OldVersion'])
                    updatePackage.newVersion = SemanticVersion(row['NewVersion'])
                    packages.append(updatePackage)
                except ValueError as ve:
                    eMsg: str = (
                        f'Package: `{packageName}` '
                        'has invalid semantic version: '
                        '{row["OldVersion"]} or {row["NewVersion"]} ignored '
                        'not updated'
                    )
                    secho(eMsg, reverse=True)

        return packages


@command(epilog=EPILOG)
@version_option(version=f'{__version__}', message='%(prog)s version %(version)s')
@option('--specification', '-s', is_flag=False, flag_value='versionSpecification.csv', default='versionSpecification.csv',
        type=Path(exists=True, path_type=PyPath),
        required=False,
        help='Update the project using a specification file')
@option('--update-packages', '-u', is_flag=True,  help='Run pip install on the packages')
def updateDependencies(specification: PyPath, update_packages: bool):
    """
    \b
    This command uses the .csv file created by createSpec.  It updates existing
    files that have dependencies in them.  It can optionally run
    `pip install --upgrade ` for each package

    \b
        - setup.py
        - config.yml
        - requirements.txt
        - pyproject.toml

    It uses the following environment variables:

    \b
        GH_TOKEN       - A personal GitHub access token necessary to read repository release information
        PROJECTS_BASE -  The local directory where the python projects are based
        PROJECT       -  The name of the project;  It should be a directory name
    """
    setUpLogging()
    vUpdate: UpdateDependencies = UpdateDependencies(specification=specification)
    vUpdate.update()
    if update_packages is True:
        packages: Packages = vUpdate.packages
        for p in packages:
            package: UpdatePackage = cast(UpdatePackage, p)
            secho(f'Updating: {package.packageName} to {package.newVersion}', reverse=True)
            runCommand(f'pip install --upgrade {package.packageName}')


if __name__ == "__main__":
    updateDependencies(['--help'])




from click import command
from click import option
from click import version_option
from click import ClickException

from versionoverlord import __version__

from versionoverlord.Common import EPILOG
from versionoverlord.Common import RepositorySlug
from versionoverlord.Common import setUpLogging

from versionoverlord.githubadapter.GitHubAdapter import GitHubAdapter
from versionoverlord.githubadapter.GitHubAdapterTypes import ReleaseTitle
from versionoverlord.githubadapter.GitHubAdapterError import GitHubAdapterError


@command(epilog=EPILOG)
@version_option(version=f'{__version__}', message='%(prog)s version %(version)s')
@option('--slug',          '-s', required=True,  help='GitHub slug')
@option('--release-title', '-r', required=True,  help='The title of the release to publish')
def publishRelease(slug: RepositorySlug, release_title: ReleaseTitle):
    """
    \b
    It uses the following environment variables:

    \b
        GH_TOKEN      – A personal GitHub access token necessary to read repository release information
        PROJECTS_BASE – The local directory where the python projects are based
        PROJECT       – The name of the project;  It should be a directory name
    """
    try:
        gitHubAdapter: GitHubAdapter = GitHubAdapter()

        gitHubAdapter.publishRelease(repositorySlug=slug, releaseTitle=release_title)

    except GitHubAdapterError as e:
        raise ClickException(message=e.message)


if __name__ == "__main__":
    setUpLogging()

    # publishRelease(['--version'])
    # publishRelease(['--help'])
    publishRelease(['--slug', 'hasii2011/TestRepository', '--release-title', 'Fake Release Name'])




from typing import Callable
from typing import Dict
from typing import List
from typing import NewType
from typing import Tuple

import logging
import logging.config

from importlib.abc import Traversable

from importlib.resources import files

from platform import platform as osPlatform

from subprocess import CompletedProcess
from subprocess import run as subProcessRun

from json import load as jsonLoad

from dataclasses import dataclass
from dataclasses import field

from semantic_version import Version as SemanticVersion

THE_GREAT_MAC_PLATFORM: str = 'macOS'

CURL_CMD:         str = 'curl'
JQ_CMD:           str = 'jq'

MAC_OS_CURL_PATH: str = f'/usr/bin/{CURL_CMD} --help'
MAC_OS_JQ_PATH:   str = f'/opt/homebrew/bin/{JQ_CMD} --version'

LINUX_OS_JQ_PATH: str = f'/usr/bin/{JQ_CMD} --version'


EPILOG:               str = 'Written by Humberto A. Sanchez II (humberto.a.sanchez.ii@gmail.com)'

ENV_PROJECTS_BASE:    str = 'PROJECTS_BASE'
ENV_PROJECT:          str = 'PROJECT'
ENV_APPLICATION_NAME: str = 'APPLICATION_NAME'
ENV_GH_TOKEN:         str = 'GH_TOKEN'

DEPENDENCIES:     str = 'dependencies'

SETUP_PY:         str = 'setup.py'
REQUIREMENTS_TXT: str = 'requirements.txt'
PYPROJECT_TOML:   str = 'pyproject.toml'
INSTALL_REQUIRES: str = 'install_requires'

CIRCLE_CI_DIRECTORY: str = '.circleci'
CIRCLE_CI_YAML:      str = 'config.yml'

SPECIFICATION_FILE:           str = 'versionSpecification.csv'
RESOURCES_PACKAGE_NAME:       str = 'versionoverlord.resources'
JSON_LOGGING_CONFIG_FILENAME: str = "loggingConfiguration.json"


EQUAL_EQUAL:           str = '=='
APPROXIMATELY_EQUAL:   str = '~='
GREATER_THAN_OR_EQUAL: str = '>='
LESS_THAN_OR_EQUAL:    str = '<='
GREATER_THAN:          str = '>'
LESS_THAN:             str = '<'

MATCH_PATTERNS: List[str] = [
    EQUAL_EQUAL, APPROXIMATELY_EQUAL, GREATER_THAN_OR_EQUAL, LESS_THAN_OR_EQUAL, GREATER_THAN, LESS_THAN
]

PYPROJECT_TOML_TRAILING_CHARACTERS: str = "\"',"


def versionFactory() -> SemanticVersion:
    return SemanticVersion('0.0.0')


PackageName = NewType('PackageName', str)


@dataclass
class UpdatePackage:
    """
    Defines the package to update
    """
    packageName: PackageName     = PackageName('')
    oldVersion:  SemanticVersion = field(default_factory=versionFactory)
    newVersion:  SemanticVersion = field(default_factory=versionFactory)


Packages                 = NewType('Packages', List[UpdatePackage])
PackageLookupType        = NewType('PackageLookupType', Dict[PackageName, UpdatePackage])
UpdateDependencyCallback = NewType('UpdateDependencyCallback', Callable[[str], str])    # type: ignore
CLISlugs                 = NewType('CLISlugs', Tuple[str])

RepositorySlug = NewType('RepositorySlug', str)


@dataclass
class AdvancedSlug:
    slug:        str = ''
    packageName: str = ''


AdvancedSlugs = NewType('AdvancedSlugs', List[AdvancedSlug])


@dataclass
class SlugVersion(AdvancedSlug):
    version:     str = ''


SlugVersions = NewType('SlugVersions', List[SlugVersion])


def setUpLogging():
    """
    """
    traversable: Traversable = files(RESOURCES_PACKAGE_NAME) / JSON_LOGGING_CONFIG_FILENAME

    # loggingConfigFilename: str = resource_filename(RESOURCES_PACKAGE_NAME, JSON_LOGGING_CONFIG_FILENAME)
    loggingConfigFilename: str = str(traversable)

    with open(loggingConfigFilename, 'r') as loggingConfigurationFile:
        configurationDictionary = jsonLoad(loggingConfigurationFile)

    logging.config.dictConfig(configurationDictionary)
    logging.logProcesses = False
    logging.logThreads = False


def extractPackageName(slug: str) -> str:
    splitSlug: List[str] = slug.split(sep='/')

    pkgName: str = splitSlug[1]
    return pkgName


def extractCLISlugs(slugs: CLISlugs) -> AdvancedSlugs:

    cliSlugs: AdvancedSlugs = AdvancedSlugs([])

    for slug in slugs:

        advancedSlug: AdvancedSlug = AdvancedSlug()
        slugPackage: List[str] = slug.split(',')
        if len(slugPackage) > 1:
            advancedSlug.slug = slugPackage[0]
            advancedSlug.packageName = slugPackage[1]
        else:
            advancedSlug.slug = slug
            advancedSlug.packageName = extractPackageName(slug)

        cliSlugs.append(advancedSlug)

    return cliSlugs


def checkJQInstalled() -> bool:
    """
    Returns: `True` if the JSON processor is installed else `False`
    """
    platform: str = osPlatform(terse=True)
    if platform.startswith(THE_GREAT_MAC_PLATFORM) is True:
        return checkInstallation(MAC_OS_JQ_PATH)
    else:
        return checkInstallation(LINUX_OS_JQ_PATH)


def checkCurlInstalled() -> bool:
    return checkInstallation(MAC_OS_CURL_PATH)


def checkInstallation(commandToCheck) -> bool:
    ans:    bool = False
    status: int  = runCommand(commandToCheck)
    if status == 0:
        ans = True

    return ans


def runCommand(programToRun: str) -> int:
    """

    Args:
        programToRun:  What must be executed

    Returns:  The status return of the executed program
    """
    completedProcess: CompletedProcess = subProcessRun([programToRun], shell=True, capture_output=True, text=True, check=False)
    return completedProcess.returncode




from logging import Logger
from logging import getLogger

from pathlib import Path

from os import linesep as osLineSep

from re import search as regExSearch
from re import Match

from semantic_version import Version as SemanticVersion

from click import ClickException

from versionoverlord.Common import AdvancedSlugs
from versionoverlord.Common import MATCH_PATTERNS
from versionoverlord.Common import PYPROJECT_TOML
from versionoverlord.Common import PYPROJECT_TOML_TRAILING_CHARACTERS
from versionoverlord.Common import REQUIREMENTS_TXT
from versionoverlord.Common import SPECIFICATION_FILE
from versionoverlord.Common import SlugVersion
from versionoverlord.Common import SlugVersions

from versionoverlord.EnvironmentBase import EnvironmentBase
from versionoverlord.githubadapter.GitHubAdapter import GitHubAdapter


class TemplateHandler(EnvironmentBase):
    """
    Creates a CSV file in the following format:

        PackageName, OldVersion,NewVersion
    """

    def __init__(self, advancedSlugs: AdvancedSlugs):

        super().__init__()

        self.logger:         Logger        = getLogger(__name__)
        self._advancedSlugs: AdvancedSlugs = advancedSlugs

        self._requirementsTxt: str = self._getRequirementsText()

    def createSpecification(self):
        print(f'Creating a specification')
        gitHubAdapter: GitHubAdapter = GitHubAdapter()

        slugVersions: SlugVersions = SlugVersions([])
        for advancedSlug in self._advancedSlugs:

            semanticVersion: SemanticVersion = gitHubAdapter.getLatestVersionNumber(advancedSlug.slug)
            slugVersion:     SlugVersion     = SlugVersion(slug=advancedSlug.slug, version=str(semanticVersion), packageName=advancedSlug.packageName)

            slugVersions.append(slugVersion)

        versionUpdateSpecification: Path = Path(SPECIFICATION_FILE)
        with versionUpdateSpecification.open(mode='w') as fd:
            fd.write(f'PackageName,OldVersion,NewVersion{osLineSep}')
            for slugVersion in slugVersions:

                oldVersion: str = self._findRequirementVersion(slugVersion.packageName)

                if oldVersion == '':
                    print(f'{slugVersion.slug} Did not find requirement')
                else:
                    fd.write(f'{slugVersion.packageName},{oldVersion},{slugVersion.version}{osLineSep}')

    def _findRequirementVersion(self, packageName: str) -> str:
        """
        Can handle requirements specifications like those specified in MATCH_PATTERNS

        For pyproject.toml we have to strip the trailing quote (single or double) and comma

        Args:
            packageName:   The package name to search for

        Returns:  A version number from the requirements text that matches the package name
                  If the requirement is not listed returns an empty string
        """
        match:            Match | None = None
        requirementValue: str          = ''
        for matchPattern in MATCH_PATTERNS:
            lookupRequirement: str = f'{packageName}{matchPattern}.*{osLineSep}'

            # TODO: This regex does not work if the requirement is the last one in the file
            # and there is no blank line at the end
            match = regExSearch(pattern=lookupRequirement, string=self._requirementsTxt)

            self.logger.info(f'{match}')
            if match is None:
                continue
            else:
                fullStr: str = (self._requirementsTxt[match.start():match.end()]).strip(osLineSep)
                splitRequirement = fullStr.split(matchPattern)
                self.logger.info(f'{splitRequirement}')
                requirementValue = splitRequirement[1]
                requirementValue = requirementValue.rstrip(PYPROJECT_TOML_TRAILING_CHARACTERS)
                break

        if match is None:
            return ''
        else:
            return requirementValue

    def _getRequirementsText(self) -> str:
        """

        Returns:  The text of either a requirements.txt or pyproject.toml
        """
        requirementsPath: Path = Path(self._projectsBase) / self._projectDirectory / REQUIREMENTS_TXT
        pyprojectPath:    Path = Path(self._projectsBase) / self._projectDirectory / PYPROJECT_TOML
        if requirementsPath.exists() is True:
            requirementsTxt: str = requirementsPath.read_text()
        elif pyprojectPath.exists() is True:
            requirementsTxt = pyprojectPath.read_text()
        else:
            raise ClickException(f'Cannot find {REQUIREMENTS_TXT} or {PYPROJECT_TOML}')

        return requirementsTxt




from typing import cast

from logging import Logger
from logging import getLogger

from collections import Counter

from os import environ as osEnvironment

from re import sub as regExSub

from datetime import date
from datetime import timedelta

from github import Github
from github import UnknownObjectException
from github import GithubException

from github.GitRelease import GitRelease
from github.PaginatedList import PaginatedList
from github.Repository import Repository
from github.Milestone import Milestone

from semantic_version import Version as SemanticVersion

from codeallybasic.Common import fixURL

from versionoverlord.Common import ENV_GH_TOKEN

from versionoverlord.Common import RepositorySlug

from versionoverlord.githubadapter.GitHubAdapterTypes import AdapterMilestone
from versionoverlord.githubadapter.GitHubAdapterTypes import AdapterRelease
from versionoverlord.githubadapter.GitHubAdapterTypes import ReleaseId
from versionoverlord.githubadapter.GitHubAdapterTypes import ReleaseTitle
from versionoverlord.githubadapter.GitHubAdapterTypes import ReleaseNumber

from versionoverlord.githubadapter.GitHubAdapterError import GitHubAdapterError

DEFAULT_MILESTONE_DUE_DATE_DELTA: int = 7
DEFAULT_MILESTONE_STATE:          str = 'open'
DEFAULT_MILESTONE_DESCRIPTION:    str = 'See the associated issues'


class GitHubAdapter:
    """
    TODO:  As more methods get added I need to stop the leakage of GitHub objects

    """
    def __init__(self):
        self.logger: Logger = getLogger(__name__)

        try:
            gitHubToken: str = osEnvironment[ENV_GH_TOKEN]
        except KeyError:
            raise GitHubAdapterError(message=f'No GitHub token specified in `{ENV_GH_TOKEN}`')

        self._github: Github = Github(gitHubToken)

    def getLatestVersionNumber(self, repositorySlug: str) -> SemanticVersion:

        try:
            repo: Repository = self._github.get_repo(repositorySlug)
            self.logger.debug(f'{repo.full_name=}')
        except UnknownObjectException:
            raise GitHubAdapterError(message=f'Unknown {repositorySlug}')

        releases: PaginatedList = repo.get_releases()

        latestReleaseVersion: SemanticVersion = SemanticVersion('0.0.0')
        for release in releases:
            gitRelease: GitRelease = cast(GitRelease, release)

            if gitRelease.draft is True:
                self.logger.warning(f'{repo.full_name} Ignore pre-release {gitRelease.tag_name}')
                continue
            releaseNumber: str = gitRelease.tag_name
            numPeriods: int = self._countPeriods(releaseNumber)
            if numPeriods < 2:
                releaseNumber = f'{releaseNumber}.0'

            releaseVersion: SemanticVersion = SemanticVersion.coerce(releaseNumber)
            self.logger.debug(f'{releaseVersion=}')
            if latestReleaseVersion < releaseVersion:
                latestReleaseVersion = releaseVersion

        return latestReleaseVersion

    def createDraftRelease(self, repositorySlug: RepositorySlug, tag: SemanticVersion, message: str) -> AdapterRelease:
        """

        Args:
            repositorySlug:   A GitHub repository slug
            tag:              The tag number
            message:          Text to put into release

        Returns:  The GitHub AdapterRelease Id

        """
        try:
            repo: Repository = self._github.get_repo(repositorySlug)
            self.logger.debug(f'{repo.full_name=}')
            releaseName: ReleaseTitle = ReleaseTitle(f'Release {tag}')

            gitRelease: GitRelease = repo.create_git_release(tag=str(tag),
                                                             name=releaseName,
                                                             message=message,
                                                             draft=True,
                                                             prerelease=False,
                                                             generate_release_notes=False)

        except UnknownObjectException:
            raise GitHubAdapterError(message=f'Unknown {repositorySlug}')

        release: AdapterRelease = AdapterRelease(
            id=ReleaseId(gitRelease.id),
            draft=gitRelease.draft,
            title=ReleaseTitle(gitRelease.title),
            body=gitRelease.body,
            tag=SemanticVersion(gitRelease.tag_name)
        )
        return release

    def createMilestone(self, repositorySlug: RepositorySlug, title: str) -> AdapterMilestone:
        try:
            repo: Repository = self._github.get_repo(repositorySlug)
            self.logger.debug(f'{repo.full_name=}')

            today: date = date.today() + timedelta(days=DEFAULT_MILESTONE_DUE_DATE_DELTA)

            milestone: Milestone = repo.create_milestone(title=title,
                                                         state=DEFAULT_MILESTONE_STATE,
                                                         description=DEFAULT_MILESTONE_DESCRIPTION,
                                                         due_on=today)
            #
            # Milestone does not provide the HTML URL;  I have to coerce one
            #
            fixedURL:   str = fixURL(milestone.url)
            coercedURL: str = regExSub(pattern=r'milestones', repl='milestone', string=fixedURL)
            adapterMilestone: AdapterMilestone = AdapterMilestone(
                releaseNumber=ReleaseNumber(milestone.number),
                title=milestone.title,
                state=milestone.state,
                description=milestone.description,
                dueDate=milestone.due_on,
                milestoneUrl=coercedURL,
            )
            return adapterMilestone

        except GithubException as ge:
            raise GitHubAdapterError(message=ge.__str__())

    def deleteMilestone(self, repositorySlug: RepositorySlug, releaseNumber: ReleaseNumber):
        """

        Args:
            repositorySlug: A GitHub repository slug
            releaseNumber:  An adapter ReleaseNumber
        """
        try:
            repo: Repository = self._github.get_repo(repositorySlug)
            self.logger.debug(f'{repo.full_name=}')

            milestone: Milestone = repo.get_milestone(number=releaseNumber)
            milestone.delete()

        except GithubException as ge:
            raise GitHubAdapterError(message=ge.__str__())

    def deleteRelease(self, repositorySlug: RepositorySlug, releaseId: int):
        """

        Args:
            repositorySlug: A GitHub repository slug
            releaseId:      A git release ID

        """
        try:
            repo: Repository = self._github.get_repo(repositorySlug)
            self.logger.debug(f'{repo.full_name=}')

            gitRelease: GitRelease = repo.get_release(id=releaseId)
            gitRelease.delete_release()

        except UnknownObjectException as e:
            # self.logger.error(f'{releaseId=} {e=}')
            raise GitHubAdapterError(message=f'Release ID not found. {e=}')

    def publishRelease(self, repositorySlug: RepositorySlug, releaseTitle: ReleaseTitle):

        try:
            repo: Repository = self._github.get_repo(repositorySlug)

            releases: PaginatedList = repo.get_releases()
            # list comprehension;  Aren't I cute . . . . . .
            matchedReleases = [r for r in releases if r.title == releaseTitle]

            if len(matchedReleases) != 1:
                raise GitHubAdapterError(f'Cannot find that release: {releaseTitle}')

            matchedRelease: GitRelease = matchedReleases[0]
            matchedRelease.update_release(name=matchedRelease.title, message=matchedRelease.body, draft=False)

        except GithubException as ge:
            raise GitHubAdapterError(message=ge.__str__())

    def _countPeriods(self, releaseNumber: str) -> int:

        cnt = Counter(list(releaseNumber))
        return cnt['.']



from click import clear
from click import command
from click import secho

from click import version_option


from versionoverlord import __version__
from versionoverlord.Common import EPILOG


@command(epilog=EPILOG)
@version_option(version=f'{__version__}', message='%(prog)s version %(version)s')
def versionOverlord():
    clear()
    secho('Commands are:')
    secho('\tvo - versionOverlord')
    secho('\tqs - querySlugs')
    secho('\tcs - createSpecification')
    secho('\tud - updateDependencies')
    secho('\tdr - draftRelease')
    secho('\tbv - bumpVersion')
    secho('\tpd - pickDependencies')
    secho('\tpr - publishRelease')




from typing import Any
from typing import Dict
from typing import List
from typing import NewType

from logging import Logger
from logging import getLogger

from os import linesep as osLineSep

from subprocess import CompletedProcess
from subprocess import run as subProcessRun

from pathlib import Path

from tomllib import load as tomlLoad

from click import command
from click import option
from click import version_option
from click import secho
from click import prompt
from click import pass_context
from click import ClickException

from click import edit as clickEdit
from click import confirm as clickConfirm

from versionoverlord import __version__

from versionoverlord.Common import CURL_CMD
from versionoverlord.Common import EPILOG
from versionoverlord.Common import JQ_CMD
from versionoverlord.Common import MATCH_PATTERNS
from versionoverlord.Common import PYPROJECT_TOML
from versionoverlord.Common import SPECIFICATION_FILE
from versionoverlord.Common import checkCurlInstalled
from versionoverlord.Common import checkJQInstalled
from versionoverlord.Common import setUpLogging

from versionoverlord.EnvironmentBase import EnvironmentBase

from versionoverlord.commands.UpdateDependencies import updateDependencies

StringList     = NewType('StringList', List[str])
StringDict     = NewType('StringDict', Dict[str, str])

TomlDict       = Dict[str, Any]
ProjectDict    = NewType('ProjectDict',    Dict[str, Any])
DependencyList = NewType('DependencyList', StringList)
OptionalDict   = NewType('OptionalDict',   Dict[str, DependencyList])

OldPackageDict = NewType('OldPackageDict', Dict[str, str])

PYPROJECT_TOML_PROJECT_KEY:             str = 'project'
PYPROJECT_TOML_DEPENDENCY_KEY:          str = 'dependencies'
PYPROJECT_TOML_OPTIONAL_DEPENDENCY_KEY: str = 'optional-dependencies'


class PickDependencies(EnvironmentBase):
    """

    """
    def __init__(self):
        super().__init__()

        self.logger: Logger = getLogger(__name__)

        directory: Path = Path(self._projectDirectory)

        self._pyProjectTomlFilePath: Path = Path(self._projectsBase) / directory / Path(PYPROJECT_TOML)

        if self._pyProjectTomlFilePath.exists() is False:
            raise ClickException(f'No such file: {self._pyProjectTomlFilePath}.')

    def pickThem(self, optional_dependencies: bool):

        with open(self._pyProjectTomlFilePath, "rb") as f:

            data:         TomlDict       = tomlLoad(f)
            project:      ProjectDict    = data[PYPROJECT_TOML_PROJECT_KEY]

            if optional_dependencies is False:
                dependencies: DependencyList = project[PYPROJECT_TOML_DEPENDENCY_KEY]
            else:
                optionalDict: OptionalDict = project[PYPROJECT_TOML_OPTIONAL_DEPENDENCY_KEY]
                projectKeys    = optionalDict.keys()

                secho('Found the following optional dependencies', reverse=True, overline=True)
                for projectKey in projectKeys:
                    secho(f'{projectKey}')

                depToUpdate: str = prompt('Enter one of the above dependencies that you wish to update', type=str)
                if depToUpdate not in projectKeys:
                    raise ClickException(f'That dependency is not valid.')
                dependencies = optionalDict[depToUpdate]

        self.logger.debug(f'{dependencies}')
        self._pickDependencies(dependencies=dependencies)

    def _getDelimiter(self, dependency: str) -> str:

        matchPattern: str = ''
        for mp in MATCH_PATTERNS:
            if mp in dependency:
                matchPattern = mp
                break

        return matchPattern

    def _getLatestVersion(self, packageName: str) -> str:
        """
        Run a pair of commands to get a pypi version for a particular
        package;  Assumes that external command existence has already been
        done

        TODO:  This code goes in Common.py
        TODO:  Remove the magic strings in the curl command

        Args:
            packageName:  The name of the package on pypi

        Returns:  The last pypi package version;  If package not found return empty string
        """

        self.logger.info(f'{packageName=}')
        checkCmd: str = (
            f"curl -s https://pypi.org/pypi/{packageName}/json | jq -r '.info.version'"
        )
        completedProcess: CompletedProcess = subProcessRun([checkCmd], shell=True, capture_output=True, text=True, check=False)
        if completedProcess.returncode == 0:
            versionDescription: str = completedProcess.stdout.strip(osLineSep)
            self.logger.debug(versionDescription)
        else:
            versionDescription = ''

        return versionDescription

    def _pickDependencies(self, dependencies: DependencyList):
        """

        Args:
            dependencies:  A dependency list

        """
        dependenciesStr: str = self._makeEditString(dependencies=dependencies)

        secho('Remove any dependencies in the editor that you do not wish to update.')
        if clickConfirm('Do you want to continue?', abort=True):
            picked: str | None = clickEdit(dependenciesStr, require_save=False)
            if picked is None:
                secho('No selections.  bye bye')
            else:
                oldPkgDict: OldPackageDict = self._makeOldPackageDictionary(dependencies=dependencies)

                self._writeTheSpecificationFile(pickedPackages=picked, oldPkgDict=oldPkgDict)

    def _makeEditString(self, dependencies: DependencyList) -> str:
        """
        Turn the pyproject.toml list into string with EOL characters that
        the developer can edit in order to pick the dependencies to update

        Args:
            dependencies: The dependency list from pyproject.toml

        Returns:  A multi-line string that can be displayed in an editor
        """
        dependenciesStr: str = ''
        for dep in dependencies:
            delimiter: str       = self._getDelimiter(dep)
            splitDep:  List[str] = dep.split(delimiter)
            dependenciesStr = f'{dependenciesStr}{splitDep[0]}{osLineSep}'

        return dependenciesStr

    def _makeOldPackageDictionary(self, dependencies: DependencyList) -> OldPackageDict:
        """

        Args:
            dependencies:  The dependency list from pyproject.toml

        Returns:  A dictionary that maps a package name to its old versio
        """

        pkgDict: OldPackageDict = OldPackageDict({})
        for dep in dependencies:
            delimiter: str       = self._getDelimiter(dep)
            splitDep:  List[str] = dep.split(delimiter)
            pkgDict[splitDep[0]] = splitDep[1]

        return pkgDict

    def _writeTheSpecificationFile(self, pickedPackages: str, oldPkgDict: OldPackageDict):
        """

        Args:
            pickedPackages:   The list of packages picked by developer
            oldPkgDict:       A dictionary that has the old versions
        """

        packageList:                List[str] = pickedPackages.split(osLineSep)
        versionUpdateSpecification: Path      = Path(SPECIFICATION_FILE)

        with versionUpdateSpecification.open(mode='w') as fd:
            fd.write(f'PackageName,OldVersion,NewVersion{osLineSep}')

            for package in packageList:
                if len(package) > 0:
                    newVersion: str = self._getLatestVersion(packageName=package)

                    specificationLine: str = f'{package},{oldPkgDict[package]},{newVersion}{osLineSep}'
                    fd.write(specificationLine)


@command(epilog=EPILOG)
@version_option(version=f'{__version__}', message='%(prog)s version %(version)s')
@option('--optional-dependencies', '-o', is_flag=True,  help='Update optional dependencies')
@pass_context
def pickDependencies(ctx, optional_dependencies: bool):
    """
    \b
    * Reads pyproject.toml and picks the dependencies from the `dependencies` section or optionally
    one of the optional dependencies
    * It displays them in an editor.
    * The developer removes dependencies he/she does not want to update.
    * This command creates the dependency csv file in the same format as the `createSpecification` command.
    * It then invokes the `updateDependencies` command to update the files.  Unlike `createSpecification`, pickDependencies queries pypi to get the module versions


    It uses the following environment variables:

    \b
        GH_TOKEN      – A personal GitHub access token necessary to read repository release information
        PROJECTS_BASE – The local directory where the python projects are based
        PROJECT       – The name of the project;  It should be a directory name
    """
    if checkJQInstalled() is False:
        raise ClickException(f'{JQ_CMD} is not installed')
    if checkCurlInstalled() is False:
        raise ClickException(f'{CURL_CMD} not installed')

    pd: PickDependencies = PickDependencies()
    pd.pickThem(optional_dependencies=optional_dependencies)

    ctx.invoke(updateDependencies)


if __name__ == "__main__":

    setUpLogging()

    # pickDependencies(['--version'])
    # pickDependencies(['--help'])
    pickDependencies(['--optional-dependencies'])




from typing import List
from typing import Optional

from logging import Logger
from logging import getLogger

from os import linesep as osLineSep

from pathlib import Path

from click import command
from click import version_option
from click import option
from click import secho
from click import prompt
from click import ClickException

from versionoverlord import __version__

from versionoverlord.Common import EPILOG
from versionoverlord.Common import setUpLogging
from versionoverlord.EnvironmentBase import EnvironmentBase
from versionoverlord.commands.TagType import TagType

STANDARD_SOURCE_DIRECTORY: Path = Path('src')
STANDARD_VERSION_FILENAME: Path = Path('_version.py')

STRIP_LF_CHAR:      str = osLineSep
STRIP_SINGLE_QUOTE: str = "' "      # Including spaces
STRIP_DOUBLE_QUOTE: str = '" '      # Including spaces

VERSION_CODE_TEMPLATE: str = "__version__: str = '{}'{}"


class BumpVersion(EnvironmentBase):
    def __init__(self, packageName: Optional['str']):
        super().__init__()

        self.logger:       Logger          = getLogger(__name__)
        self._packageName: Optional['str'] = packageName

        directory:             Path = Path(self._projectDirectory)
        if packageName is None:
            packagePath: Path = directory
        else:
            packagePath = Path(packageName)
        self._versionFilePath: Path = Path(self._projectsBase) / directory / STANDARD_SOURCE_DIRECTORY / packagePath / STANDARD_VERSION_FILENAME

        if self._versionFilePath.exists() is False:
            raise ClickException(f'No such file: {self._versionFilePath}.  Perhaps specify package name.')

    def bumpIt(self):

        versionCode:  str       = self._versionFilePath.read_text()
        versionSplit: List[str] = versionCode.split('=')
        versionStr:   str       = versionSplit[1]

        versionStr = versionStr.strip(STRIP_LF_CHAR).strip(STRIP_SINGLE_QUOTE).strip(STRIP_DOUBLE_QUOTE)
        secho(f'Current version: `{versionStr}`')

        newVersion: TagType = prompt('Enter new version', type=TagType())
        secho(f'New version: {newVersion}')
        updateCodeLine: str = VERSION_CODE_TEMPLATE.format(newVersion, osLineSep)
        self._versionFilePath.write_text(updateCodeLine)


@command(epilog=EPILOG)
@version_option(version=f'{__version__}', message='%(prog)s version %(version)s')
@option('--package-name', '-p', help='Optional package name when repository does not match')
def bumpVersion(package_name: str):
    """
    \b
    Bump version looks for file in src/<moduleName>/_version.py.  It echoes it to stdout
    and asks the developer to provide an updated value.

    It uses the following environment variables:

    \b
        GH_TOKEN      – A personal GitHub access token necessary to read repository release information
        PROJECTS_BASE – The local directory where the python projects are based
        PROJECT       – The name of the project;  It should be a directory name
    """
    bv: BumpVersion = BumpVersion(packageName=package_name)
    bv.bumpIt()


if __name__ == "__main__":

    setUpLogging()
    # noinspection SpellCheckingInspection
    # createSpecification(['-i', 'tests/resources/testdata/query.slg'])
    # createSpecification(['-s', 'hasii2011/code-ally-basic,codeallybasic'])

    # bumpVersion(['--version'])
    # bumpVersion(['--help'])
    bumpVersion([])



from typing import cast

from click import command
from click import version_option
from click import option
from click import ClickException

from semantic_version import Version as SemanticVersion

from versionoverlord import __version__
from versionoverlord.Common import EPILOG
from versionoverlord.Common import RepositorySlug

from versionoverlord.Common import setUpLogging
from versionoverlord.githubadapter.GitHubAdapter import GitHubAdapter
from versionoverlord.commands.TagType import TagType
from versionoverlord.githubadapter.GitHubAdapterTypes import AdapterMilestone
from versionoverlord.githubadapter.GitHubAdapterError import GitHubAdapterError

RELEASE_STUB_MESSAGE_TEMPLATE: str = 'See issues associated with this [milestone]({})'


@command(epilog=EPILOG)
@version_option(version=f'{__version__}', message='%(prog)s version %(version)s')
@option('--slug',      '-s', required=True,                 help='GitHub slug')
@option('--tag',       '-t', required=True, type=TagType(), help='Tag for release as a semantic version')
@option('--milestone', '-m', is_flag=True,                  help='Create associated milestone')
def draftRelease(slug: RepositorySlug, tag: TagType, milestone: bool):
    """
    \b
    This command creates draft release in the appropriate repository.
    You must provide a repository slug.
    The tag is a string that complies with the Semantic Version specification
    Specify the milestone if you want to created an associated milestone

    It uses the following environment variables:

    \b
        GH_TOKEN      – A personal GitHub access token necessary to read repository release information
        PROJECTS_BASE – The local directory where the python projects are based
        PROJECT       – The name of the project;  It should be a directory name
    """
    try:
        gitHubAdapter: GitHubAdapter = GitHubAdapter()
        milestoneUrl: str = ''
        if milestone is True:
            milestoneTitle: str = f'Release {tag}'
            adapterMilestone: AdapterMilestone = gitHubAdapter.createMilestone(repositorySlug=slug, title=milestoneTitle)
            milestoneUrl = adapterMilestone.milestoneUrl

        message: str = RELEASE_STUB_MESSAGE_TEMPLATE.format(milestoneUrl)
        gitHubAdapter.createDraftRelease(repositorySlug=slug, tag=cast(SemanticVersion, tag), message=message)
    except GitHubAdapterError as e:
        raise ClickException(message=e.message)


if __name__ == "__main__":
    setUpLogging()
    # noinspection SpellCheckingInspection
    # createSpecification(['-i', 'tests/resources/testdata/query.slg'])
    # createSpecification(['-s', 'hasii2011/code-ally-basic,codeallybasic'])
    # -s hasii2011/ -s hasii2011/buildlackey

    # draftRelease(['--version'])
    # draftRelease(['--help'])
    draftRelease(['--slug', 'hasii2011/TestRepository', '--tag', '10.0.0', '--milestone'])
